@@ -1,26 +1,31 @@
+
 import os
 import sys
-if sys.version_info[0] >= 3:
-    from urllib.request import urlopen
-    from urllib.parse import quote_plus
-    from urllib.parse import urlencode
-    from urllib.error import HTTPError
-else:
-    from urllib2 import urlopen
-    from urllib2 import HTTPError
-    from urllib import quote_plus
-    from urllib import urlencode
-
 import xml.etree.ElementTree as ET
-from dateutil.parser import parse
+if sys.version_info[0] >= 3:
+    import urllib.request as url_request
+    import urllib.parse as url_parse
+    import urllib.error as url_error
+else:
+    import urllib2 as url_request
+    import urllib as url_parse
+    import urllib2 as url_error
+
 import pandas as pd
 
+urlopen = url_request.urlopen
+quote_plus = url_parse.quote_plus
+urlencode = url_parse.urlencode
+HTTPError = url_error.HTTPError
+
 
 class Fred(object):
     earliest_realtime_start = '1776-07-04'
     latest_realtime_end = '9999-12-31'
     nan_char = '.'
     max_results_per_request = 1000
+    root_url = 'https://api.stlouisfed.org/fred'
+
     def __init__(self,
                  api_key=None,
                  api_key_file=None):
@@ -41,15 +46,20 @@ class Fred(object):
             self.api_key = os.environ.get('FRED_API_KEY')
 
         if self.api_key is None:
-            raise ValueError("You need to set a valid API key. You can set it in 3 ways: pass the string with api_key, "
-                             "or set api_key_file to a file with the api key in the first line, or set the environment "
-                             "variable 'FRED_API_KEY' to the value of your api key. You can sign up for a free api key "
-                             "on the Fred website at http://research.stlouisfed.org/fred2/")
+            import textwrap
+            raise ValueError(textwrap.dedent("""\
+                    You need to set a valid API key. You can set it in 3 ways:
+                    pass the string with api_key, or set api_key_file to a
+                    file with the api key in the first line, or set the
+                    environment variable 'FRED_API_KEY' to the value of your
+                    api key. You can sign up for a free api key on the Fred
+                    website at http://research.stlouisfed.org/fred2/"""))
 
     def __fetch_data(self, url):
         """
         helper function for fetching data given a request URL
         """
+        url += '&api_key=' + self.api_key
         try:
             response = urlopen(url)
             root = ET.fromstring(response.read())
@@ -58,6 +68,15 @@ class Fred(object):
             raise ValueError(root.get('message'))
         return root
 
+    def _parse(self, date_str, format='%Y-%m-%d'):
+        """
+        helper function for parsing FRED date string into datetime
+        """
+        rv = pd.to_datetime(date_str, format=format)
+        if hasattr(rv, 'to_pydatetime'):
+            rv = rv.to_pydatetime()
+        return rv
+
     def get_series_info(self, series_id):
         """
         Get information about a series such as its title, frequency, observation start/end dates, units, notes, etc.
@@ -72,11 +91,11 @@ class Fred(object):
         info : Series
             a pandas Series containing information about the Fred series
         """
-        url = "http://api.stlouisfed.org/fred/series?series_id=%s&api_key=%s" % (series_id, self.api_key)
+        url = "%s/series?series_id=%s" % (self.root_url, series_id)
         root = self.__fetch_data(url)
-        if root is None:
+        if root is None or not len(root):
             raise ValueError('No info exists for series id: ' + series_id)
-        info = pd.Series(root.getchildren()[0].attrib)
+        info = pd.Series(list(root)[0].attrib)
         return info
 
     def get_series(self, series_id, observation_start=None, observation_end=None, **kwargs):
@@ -92,36 +111,34 @@ class Fred(object):
         observation_end : datetime or datetime-like str such as '7/1/2014', optional
             latest observation date
         kwargs : additional parameters
-            Any additional parameters supported by FRED. You can see http://api.stlouisfed.org/docs/fred/series_observations.html for the full list
+            Any additional parameters supported by FRED. You can see https://api.stlouisfed.org/docs/fred/series_observations.html for the full list
 
         Returns
         -------
         data : Series
             a Series where each index is the observation date and the value is the data for the Fred series
         """
-        url = "http://api.stlouisfed.org/fred/series/observations?series_id=%s&api_key=%s" % (series_id, self.api_key)
-
+        url = "%s/series/observations?series_id=%s" % (self.root_url, series_id)
         if observation_start is not None:
-            observation_start = pd.to_datetime(observation_start, errors='raise')
+            observation_start = pd.to_datetime(observation_start,
+                                               errors='raise')
             url += '&observation_start=' + observation_start.strftime('%Y-%m-%d')
         if observation_end is not None:
             observation_end = pd.to_datetime(observation_end, errors='raise')
             url += '&observation_end=' + observation_end.strftime('%Y-%m-%d')
-
-        if kwargs is not None:
+        if kwargs.keys():
             url += '&' + urlencode(kwargs)
-
         root = self.__fetch_data(url)
         if root is None:
             raise ValueError('No data exists for series id: ' + series_id)
         data = {}
-        for child in root.getchildren():
+        for child in root:
             val = child.get('value')
             if val == self.nan_char:
                 val = float('NaN')
             else:
                 val = float(val)
-            data[parse(child.get('date'))] = val
+            data[self._parse(child.get('date'))] = val
         return pd.Series(data)
 
     def get_series_latest_release(self, series_id):
@@ -203,24 +220,24 @@ class Fred(object):
             a DataFrame with columns 'date', 'realtime_start' and 'value' where 'date' is the observation period and 'realtime_start'
             is when the corresponding value (either first release or revision) is reported.
         """
-        url = "http://api.stlouisfed.org/fred/series/observations?series_id=%s&api_key=%s&realtime_start=%s&realtime_end=%s" % (series_id,
-                                                                                                                                self.api_key,
-                                                                                                                                self.earliest_realtime_start,
-                                                                                                                                self.latest_realtime_end)
+        url = "%s/series/observations?series_id=%s&realtime_start=%s&realtime_end=%s" % (self.root_url,
+                                                                                         series_id,
+                                                                                         self.earliest_realtime_start,
+                                                                                         self.latest_realtime_end)
         root = self.__fetch_data(url)
         if root is None:
             raise ValueError('No data exists for series id: ' + series_id)
         data = {}
         i = 0
-        for child in root.getchildren():
+        for child in root:
             val = child.get('value')
             if val == self.nan_char:
                 val = float('NaN')
             else:
                 val = float(val)
-            realtime_start = parse(child.get('realtime_start'))
-            # realtime_end = parse(child.get('realtime_end'))
-            date = parse(child.get('date'))
+            realtime_start = self._parse(child.get('realtime_start'))
+            # realtime_end = self._parse(child.get('realtime_end'))
+            date = self._parse(child.get('date'))
 
             data[i] = {'realtime_start': realtime_start,
                        # 'realtime_end': realtime_end,
@@ -245,13 +262,13 @@ class Fred(object):
         dates : list
             list of vintage dates
         """
-        url = "http://api.stlouisfed.org/fred/series/vintagedates?series_id=%s&api_key=%s" % (series_id, self.api_key)
+        url = "%s/series/vintagedates?series_id=%s" % (self.root_url, series_id)
         root = self.__fetch_data(url)
         if root is None:
             raise ValueError('No vintage date exists for series id: ' + series_id)
         dates = []
-        for child in root.getchildren():
-            dates.append(parse(child.text))
+        for child in root:
+            dates.append(self._parse(child.text))
         return dates
 
     def __do_series_search(self, url):
@@ -265,7 +282,7 @@ class Fred(object):
 
         num_results_returned = 0  # number of results returned in this HTTP request
         num_results_total = int(root.get('count'))  # total number of results, this can be larger than number of results returned
-        for child in root.getchildren():
+        for child in root:
             num_results_returned += 1
             series_id = child.get('id')
             series_ids.append(series_id)
@@ -280,14 +297,14 @@ class Fred(object):
             data = pd.DataFrame(data, columns=series_ids).T
             # parse datetime columns
             for field in ["realtime_start", "realtime_end", "observation_start", "observation_end", "last_updated"]:
-                data[field] = data[field].apply(parse)
+                data[field] = data[field].apply(self._parse, format=None)
             # set index name
             data.index.name = 'series id'
         else:
             data = None
         return data, num_results_total
 
-    def __get_search_results(self, url, limit, order_by, sort_order):
+    def __get_search_results(self, url, limit, order_by, sort_order, filter):
         """
         helper function for getting search results up to specified limit on the number of results. The Fred HTTP API
         truncates to 1000 results per request, so this may issue multiple HTTP requests to obtain more available data.
@@ -302,6 +319,12 @@ class Fred(object):
             else:
                 raise ValueError('%s is not in the valid list of order_by options: %s' % (order_by, str(order_by_options)))
 
+        if filter is not None:
+            if len(filter) == 2:
+                url = url + '&filter_variable=%s&filter_value=%s' % (filter[0], filter[1])
+            else:
+                raise ValueError('Filter should be a 2 item tuple like (filter_variable, filter_value)')
+
         sort_order_options = ['asc', 'desc']
         if sort_order is not None:
             if sort_order in sort_order_options:
@@ -325,7 +348,7 @@ class Fred(object):
                 data = data.append(next_data)
         return data.head(max_results_needed)
 
-    def search(self, text, limit=1000, order_by=None, sort_order=None):
+    def search(self, text, limit=1000, order_by=None, sort_order=None, filter=None):
         """
         Do a fulltext search for series in the Fred dataset. Returns information about matching series in a DataFrame.
 
@@ -341,17 +364,21 @@ class Fred(object):
             'popularity'
         sort_order : str, optional
             sort the results by ascending or descending order. Valid options are 'asc' or 'desc'
+        filter : tuple, optional
+            filters the results. Expects a tuple like (filter_variable, filter_value).
+            Valid filter_variable values are 'frequency', 'units', and 'seasonal_adjustment'
 
         Returns
         -------
         info : DataFrame
             a DataFrame containing information about the matching Fred series
         """
-        url = "http://api.stlouisfed.org/fred/series/search?search_text=%s&api_key=%s" % (quote_plus(text), self.api_key)
-        info = self.__get_search_results(url, limit, order_by, sort_order)
+        url = "%s/series/search?search_text=%s&" % (self.root_url,
+                                                    quote_plus(text))
+        info = self.__get_search_results(url, limit, order_by, sort_order, filter)
         return info
 
-    def search_by_release(self, release_id, limit=0, order_by=None, sort_order=None):
+    def search_by_release(self, release_id, limit=0, order_by=None, sort_order=None, filter=None):
         """
         Search for series that belongs to a release id. Returns information about matching series in a DataFrame.
 
@@ -367,19 +394,22 @@ class Fred(object):
             'popularity'
         sort_order : str, optional
             sort the results by ascending or descending order. Valid options are 'asc' or 'desc'
+        filter : tuple, optional
+            filters the results. Expects a tuple like (filter_variable, filter_value).
+            Valid filter_variable values are 'frequency', 'units', and 'seasonal_adjustment'
 
         Returns
         -------
         info : DataFrame
             a DataFrame containing information about the matching Fred series
         """
-        url = "http://api.stlouisfed.org/fred/release/series?release_id=%d&&api_key=%s" % (release_id, self.api_key)
-        info = self.__get_search_results(url, limit, order_by, sort_order)
+        url = "%s/release/series?release_id=%d" % (self.root_url, release_id)
+        info = self.__get_search_results(url, limit, order_by, sort_order, filter)
         if info is None:
             raise ValueError('No series exists for release id: ' + str(release_id))
         return info
 
-    def search_by_category(self, category_id, limit=0, order_by=None, sort_order=None):
+    def search_by_category(self, category_id, limit=0, order_by=None, sort_order=None, filter=None):
         """
         Search for series that belongs to a category id. Returns information about matching series in a DataFrame.
 
@@ -395,14 +425,18 @@ class Fred(object):
             'popularity'
         sort_order : str, optional
             sort the results by ascending or descending order. Valid options are 'asc' or 'desc'
+        filter : tuple, optional
+            filters the results. Expects a tuple like (filter_variable, filter_value).
+            Valid filter_variable values are 'frequency', 'units', and 'seasonal_adjustment'
 
         Returns
         -------
         info : DataFrame
             a DataFrame containing information about the matching Fred series
         """
-        url = "http://api.stlouisfed.org/fred/category/series?category_id=%d&api_key=%s" % (category_id, self.api_key)
-        info = self.__get_search_results(url, limit, order_by, sort_order)
+        url = "%s/category/series?category_id=%d&" % (self.root_url,
+                                                      category_id)
+        info = self.__get_search_results(url, limit, order_by, sort_order, filter)
         if info is None:
             raise ValueError('No series exists for category id: ' + str(category_id))
         return info
